"""
2021, Babbage B.V.
Extension to the USA add-on to let USA follow spoken words during SayAll
"""

import sys
import os

import speech
import ui
from speech import sayAll
from logHandler import log
from vision.constants import Context
import api
import textInfos
from . import textProcessor, newSpeech, rectHighlighter, textDifference
import NVDAObjects




class wordFollower:
    """
    General entrance to the functionality of following words while reading text
    """

    def log(self, string : str):
        log.debugWarning(f"Wordfollower USA: {string}")

    def __init__(self, USAinstance):
        self.USA = USAinstance
        self.enabled = False
        self.rectHighlighter = None

        if self.USA.getSettings().followWord:
            self.enable_wordFollower(notify=False)

    def terminate(self):
        self.log("terminating wordFollower")
        # disable itself if enabled
        if self.enabled:
            self.disable_wordFollower(notify=False)
        if self.rectHighlighter:
            self.rectHighlighter.terminate()
        self.log("wordFollower terminated")


    def enable_wordFollower(self, notify = True):
        self.log("USA: enable_wordFollower")

        sayAll.SayAllHandler._getTextInfoSpeech = newSpeech.getTextInfoSpeech

        # overwrite function that processes text sequence
        newSpeech.adaptRelativeSpeechSequence = self.adaptRelativeSpeechSequence

        if self.USA.getSettings().followWordRect:
            self.enable_wordFollowerRect(notify=False)

        self.enabled = True
        if notify:
            ui.message(_('Word follower for USA enabled'))

    def handleForegroundChange(self, obj):
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(None)

    def handleReviewMove(self, context):
        pass

    def handleFocusChange(self, obj):
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(None)

    def disable_wordFollower(self, notify = True):
        log.info("USA: disable_wordFollower")
        self.disable_wordFollowerRect(notify=False)
        # reset overwritten functions
        sayAll.SayAllHandler._getTextInfoSpeech = speech.getTextInfoSpeech

        self.enabled = False
        if notify:
            ui.message(_('Word follower for USA disabled'))

    def enable_wordFollowerRect(self, notify = True):
        log.info("USA: enable_wordFollowerRect")
        if self.USA.getSettings().followWord:
            self.rectHighlighter = rectHighlighter.rectHighlighter(self.USA)
            if notify:
                ui.message(_('Word follower rectangle for USA enabled'))

    def disable_wordFollowerRect(self, notify = True):
        log.info("USA: disable_wordFollowerRect")
        if self.rectHighlighter:
            self.rectHighlighter.terminate()
            self.rectHighlighter = None
            if notify:
                ui.message(_('Word follower rectangle for USA disabled'))

    def adaptRelativeSpeechSequence(self, sequence, offset = 0):
        # adapts original speech sequence to add callback commands in them.
        # This function only checks whether there is an sayall reader active. If it is, it passes the sequence to
        # the wordSplitter function

        try:
            # get the current textInfo that is being read
            if sayAll.SayAllHandler._getActiveSayAll():
                info = sayAll.SayAllHandler._getActiveSayAll().reader
            else:
                log.error(f"SPEECH: sayAllHandler does not have _activeSayAll()")
                return sequence

            # split the sequence into words with callback commands to the location
            newSequence = textProcessor.wordSplitter(info, sequence, self.onCallBack, offset)
            return newSequence

        except Exception as e:
            # extract error message for logging, return default sequence
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            locationString = f"{(exc_type, fname, exc_tb.tb_lineno)}"

            log.error(f"SPEECH: Something went wrong, just return sequence as regular sayAll. \n location: {locationString} \n Error {e}")
            return sequence

    def testFunction(self, value):
        self.log(f"received value! {value}")

    def correctForMissalignment(self, reader, originalText, bookmark):
        """
        In vbufs there is a possibility that the text changes between the making of the callback and the calling of the
        callback. If that occurs the offsets previously calculated do not work anymore. A function is called that looks
        at the difference between the original text and the current text and tries to deduct what position it should
        be in.
        """
        self.log("correction for potential misallignment")
        # in virtual Buffers the text can change and thus the location of original word that should be spoken.
        # we check what is different between the original and current text and correct the bookmark for it.

        # get the current text
        reader.move(textInfos.UNIT_STORY, direction=-1, endPoint='start')
        reader.move(textInfos.UNIT_STORY, direction=1, endPoint='end')

        # get list with all matches of the current and original text

        deltaList = textDifference.getDelta(reader.text, originalText)

        correction = 0
        for ii in deltaList:
            # the correction is the difference in offset up to the position.
            position = ii[0]
            correction = ii[1]
            if bookmark.startOffset < position:
                # the positions of deltaList are in order. if our bookmark is below a position we have the correct correction
                break

        # add the correction to the list, correction = 0 if no correction was found
        reader._startOffset = bookmark.startOffset + correction
        reader._endOffset = bookmark.endOffset + correction
        return reader

    def onCallBack(self, ii):
        """
        function used as callBackCommand that is injected into the speech sequence.
        bookmark: holds the location of the word
        word: the word that is being read
        originalText: the coarse text that was on the page when the callback command was created
        reader: original textInfo used by the sayAll handler
        """
        word, bookmark, originalText, reader = textProcessor.sayAllWordDataBase[ii]
        log.info(f"{repr(word)}")
        del textProcessor.sayAllWordDataBase[ii]

        if reader:
            # set reader to bookmark offset
            reader = textProcessor.moveStartEndOffsetsByMove(reader, bookmark)

            # does not work with UIA (yet)
            if not reader.text == word:
                reader = self.correctForMissalignment(reader, originalText, bookmark)

            if len(word) and word[-1] == " ":
                word = word[:-1]
                reader.move(textInfos.UNIT_CHARACTER, -1, endPoint="end")
            # get the rectangle of the text if possible, and optionally scroll towards it
            rect = textProcessor.getBoundingRects(reader)
            #log.info(f"{word} {rect} {reader.boundingRects} {reader} {reader.text} {reader.obj} {reader.obj.states}")
            # update the caret if applicable.
            if self.USA.getSettings().followWordJumpCaret:
                reader.collapse()
                reader.updateCaret()

            # highlight the word if applicable.
            if self.rectHighlighter:
                self.rectHighlighter.updateCustomRect(rect)

            # track to the rectangle
            self.USA.trackToRectangle(rect, context=Context.CARET)
        else:
            # we could not get the reader anymore
            log.error("SayAll does currently not have a reader object but we are here!")
            if self.rectHighlighter:
                self.rectHighlighter.updateCustomRect(None)



