"""
Babbage B.V. 2019-2023

visionEnhancementProviders/usa.py
Contains the USA class. Also implements some functions for zooming in/out with the scroll-wheel and NVDA+mouseclick
"""

from ._base import USABase, libUSA, debug
from ._base.dialogs import USASettingsPanel
from ._base.settings import USASettings
from ._base.visionUtilCopy import *
from scriptHandler import script
import ui
from logHandler import log
from locationHelper import RectLTWH, RectLTRB
import addonHandler
import globalVars
import mouseHandler
import time
import core
import winVersion
import speech
import watchdog
import ctypes
import ctypes.wintypes
import keyboardHandler, winInputHook
addonHandler.initTranslation()

import api
import globalCommands

import wx
import gui
from ._base.wordFollower import objectReader, wordFollower
import browseMode


oldMouseHook = None # function that is the mouse hook of NVDA itself. This overwritten by USA
def is_key_pressed(virtual_key_code):
	return bool(ctypes.windll.user32.GetKeyState(virtual_key_code) >> 15)

class USA(USABase):
	_settings = USASettings()

	@classmethod
	def getSettingsPanelClass(cls):
		return USASettingsPanel

	@classmethod
	def canStart(cls) -> bool:
		return winVersion.isFullScreenMagnificationAvailable()

	def __init__(self):
		log.info("starting USA")

		if self.getSettings().isRemote:
			pass
		else:
			self.c_onTrial = libUSA.LicensingCallbackFuncType(self._onTrial)
			self.c_onTrialExpired = libUSA.LicensingCallbackFuncType(self._onTrialExpired)
			libUSA.SetCallbacks(self.c_onTrial, self.c_onTrialExpired)
			self._hostWindowHandle = libUSA.Initialize(libUSA.MAG_TYPE_SCREEN)

		self.hookOurMouseFunction()
		# try:
		#     hookThread = threading.Thread(
		#         name=__name__,  # winInputHook
		#         target=self.hookThreadFunc2
		#     )
		#     hookThread.start()
		#     # overwriting mouse event for scroll zooming
		#     log.info("USA: starting mouse hook")
		#     log.info("USA: plugin initialized")
		# except:
		#     log.error("could not initialize threading for mouse hook for scrolling zoom in or out")

		super().__init__()
		self.objectReader = objectReader.objectReader(self)
		self.wordFollower = wordFollower.wordFollower(self)

	def terminate(self):
		log.info("terminating USA")

		self.unhookOurMouseFunction()
		self.wordFollower.terminate()
		self.objectReader.terminate()
		try:
			super().terminate()
		finally:
			try:
				libUSA.Terminate()
			finally:
				libUSA.SetCallbacks(None, None)
		log.info("USA terminated")



	def secondCallToCatchLicenseMistakes(self):
		if not libUSA.IsLicensed():
			wx.CallAfter(
				wx.CallLater,
				1000,
				gui.messageBox,
				parent=gui.mainFrame,
				# Trannlators: Title of the trial dialog.
				caption=_("Trial"),
				# Translators: Message shown when running in trial mode.
				message=_(
					"Ultimate Screen Access is running in 10 minutes trial mode.\n"
					"If this message is unexpected, please check whether you have a valid license.\n"
				),
				style=wx.OK | wx.ICON_WARNING
			)
		else:
			log.info("USA licensed on second try!")

	def _onTrial(self):
		if not globalVars.appArgs.secure:
			log.error("USA on trial! call libUSA to check for license again to be sure")
			wx.CallAfter(
				wx.CallLater,
				100,
				libUSA.IsLicensed()

			)
			wx.CallAfter(
				wx.CallLater,
				2000,
				self.secondCallToCatchLicenseMistakes
			)





	def _onTrialExpired(self):
		if not globalVars.appArgs.secure:
			wx.CallAfter(
				gui.messageBox,
				parent=gui.mainFrame,
				# Translators: Title of the trial expireddialog.
				caption=_("Trial Expired"),
				# Translators: Message shown when trial is expired.
				message=_("Your 10 minutes trial of Ultimate Screen Access has expired."),
				style=wx.OK | wx.ICON_WARNING
			)
		self.unregisterEventExtensionPoints()
		core.callLater(100, self.terminate)

	def handleForegroundChange(self, obj):
		log.debug(f"USA: handleForegroundChange {obj}")
		self.wordFollower.handleForegroundChange(obj)
		self.objectReader.handleForegroundChange(obj)

		# for the focus dependent color schemes we need to change the color scheme if the appName is in the custom made
		# aggerated section of conf
		appName = api.getForegroundObject().appModule.appName
		if self.getSettings().conf['fgColorschemes'].get(appName):
			self.getSettings()._set_colorScheme(self.getSettings().conf['fgColorschemes'].get(appName))

		super().handleForegroundChange(obj)

	def handleReviewMove(self, context):
		log.debug(f"USA: handleReviewMove {context}")
		self.wordFollower.handleReviewMove(context)
		self.objectReader.handleReviewMove(context)
		super().handleReviewMove(context)

	def handleFocusChange(self, obj):
		log.debug(f"USA: handleFocusChange {obj}")
		self.wordFollower.handleFocusChange(obj)
		self.objectReader.handleFocusChange(obj)
		super().handleFocusChange(obj)

	@script(gesture="kb:alt+NVDA+upArrow", description=_("Scroll magnification area up"))
	def script_scrollUp(self, gesture):
		self._scrollScriptHelper( yDir=-1)

	@script(gesture="kb:alt+NVDA+downArrow", description=_("Scroll magnification area down"))
	def script_scrollDown(self, gesture):
		self._scrollScriptHelper( yDir=1)

	@script(gesture="kb:alt+NVDA+leftArrow", description=_("Scroll magnification area left"))
	def script_scrollLeft(self, gesture):
		self._scrollScriptHelper( xDir=-1)

	@script(gesture="kb:alt+NVDA+rightArrow", description=_("Scroll magnification area right"))
	def script_scrollRight(self, gesture):
		self._scrollScriptHelper( xDir=1)

	@script(
		gesture = "kb:NVDA+alt+r",
		description = _("Start reading all from object beneath the mouse")
	)
	def script_readAtMousePos(self, gesture):
		log.debugWarning(f"USA: script_readAtMousePos \n \t self.objectReader.enabled = {self.objectReader.enabled}")
		if self.objectReader.enabled:
			self.objectReader.readAllUnderMouse()

	@script(
		description= _("Toggles word follower of USA")
	)
	def script_toggleFollower(self, gesture):
		# script to toggle on/off the word follower
		settings = self.getSettings()
		conf = settings.conf
		if settings.followWord:
			self.wordFollower.disable_wordFollower()
		else:
			self.wordFollower.enable_wordFollower()
		followWord = not settings.followWord
		settings._set_followWord(followWord)
		conf['followWord'] = followWord
		debug(f"USA: script_toggleFollower \n \t settings.followWord = {settings.followWord}")

	@script(
		description= _("Toggles rectangle for word follower of USA")
	)
	def script_toggleFollowerRect(self, gesture):
		# script to toggle on/off the rectangle for the word follower, only changes if wordfollower itself is enabled
		settings = self.getSettings()
		conf = settings.conf
		if settings.followWord:
			if settings.followWordRect:
				self.wordFollower.disable_wordFollowerRect()
			else:
				self.wordFollower.enable_wordFollowerRect()
			followWordRect = not settings.followWordRect
			settings._set_followWordRect(followWordRect)
			conf['followWordRect'] = followWordRect

		debug(f"USA: script_toggleFollowerRect \n \t settings.followWord = {settings.followWord} \n \t settings.followWordRect = {settings.followWordRect}")

	@script(
		gesture = "kb:NVDA+control+p",
		description= _("Set current color scheme as default for current application")
	)
	def script_setFGColorScheme(self, gesture):
		debug(f"USA: script_setFGColorScheme \n \t appName = {api.getForegroundObject().appModule.appName} \n \t colorScheme = {self.getSettings().colorScheme}")
		# save the current color scheme for the current fg application, which is then used for the focus dependent color
		# scheme feature
		appName = api.getForegroundObject().appModule.appName
		self.getSettings().conf['fgColorschemes'][appName] = self.getSettings().colorScheme
		ui.message(_("set current color scheme for this app as default"))


	def getUnicornCommandMapping(self):
		mapping = {"toggleColorScheme": self.toggleColorScheme,
				   "trackToRectangle": self.trackToRectangle,
				   "reportOffsets": self.reportOffsets,
				   "toggleBitmapSmoothing": self.toggleBitmapSmoothing,
				   "magnificationLevelHelper": self._setMagnificationLevelHelper,
				   "scrollScriptHelper": self._scrollScriptHelper,
				   "toggleMagnifier": self.toggleMagnifier}
		return mapping

	def deserializeUnicornArguments(self,command, **kwargs):
		if command == "trackToRectangle":
			context = kwargs['context']
			log.info(f" {kwargs['rect'] } ")
			rect = RectLTRB(*kwargs['rect'])
			return {'rect': rect, 'context': context}
		elif command == "scrollScriptHelper":
			xDir = kwargs['xDir']
			yDir = kwargs['yDir']
			return {'xDir': xDir, 'yDir': yDir}
		elif command == "magnificationLevelHelper":
			increase = kwargs['increase']
			return {'increase': increase}


	def toggleColorScheme(self):
		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="toggleColorScheme")
		else:
			settings = self.getSettings()
			transformations = list(settings.availableColorschemes.values())
			IDs = list(transformation.id for transformation in transformations)
			try:
				index = IDs.index(settings.colorScheme)
			except ValueError:
				index=0
			newIndex = (index + 1) % len(IDs)
			new = transformations[newIndex]
			settings.colorScheme = IDs[newIndex]
			# Translators: Reports the newly activated color scheme.

			ui.message(_("Color scheme {name}").format(name=new.displayName))

	@script(gestures=["kb:alt+NVDA+backspace", "kb:control+backspace"], description=_('Cycles between the available color schemes'))
	def script_toggleColorScheme(self,gesture):
		self.toggleColorScheme()

	def trackToRectangle(self, rect, context=""):
		if self.getSettings().isRemote:
			if isinstance(rect, locationHelper.RectLTWH): rect = rect.toLTRB()
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="trackToRectangle", rect=rect, context=context)
		else:
			log.debug(f"USA: trackToRectangle {rect} {context}")
			if not rect:
				log.warning("No rectangle, unexpected behavior")
				return
			settings = self.getSettings()
			if (
				context == Context.FOCUS
				and settings.trackToMouse
				and (time.time() - mouseHandler.lastMouseEventTime) * 1000 < settings.conf['mouseFocusTimeout']
			):
				# Avoid display bouncing while the mouse is being used.
				return
			r = rect.toRECT()
			boundary = self.getContextBoundary(context)
			for hwnd in libUSA.magWindows:
				libUSA.SetTrackingRectangle(hwnd, r, boundary)

	def reportOffsets(self):
		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="reportOffsets")
		else:
			point = libUSA.GetTrackingOffsets(None)
			left = point.x
			top = point.y
			ui.message(str((left, top)))

	@script(gesture="kb:NVDA+i")
	def script_reportOffsets(self, gesture):
		self.reportOffsets()

	def toggleBitmapSmoothing(self):
		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn( command="toggleBitmapSmoothing")
		else:
			settings = self.getSettings()
			oldState = settings.bitmapSmoothing
			settings.bitmapSmoothing = not oldState
			if settings.bitmapSmoothing == oldState:
				ui.message(_("Bitmap smoothing not available"))
				return
			stateMsg = _("on") if settings.bitmapSmoothing else _("off")
			ui.message(_("Bitmap smoothing {state}").format(state=stateMsg))

	@script(gesture="kb:alt+NVDA+s", description=_('Toggles the state of bitmap smoothing'))
	def script_toggleBitmapSmoothing(self, gesture):
		self.toggleBitmapSmoothing()

	def _setMagnificationLevelHelper(self, increase: bool = True):
		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="magnificationLevelHelper", increase = increase)
		else:
			level = self.getSettings()._setMagnificationLevelHelper(increase=increase)
			speech.cancelSpeech()
			levelString = "{:.1f}".format(level)
			ui.message(_("Magnification level {level}").format(level=levelString))

	def _scrollScriptHelper(self, xDir=0, yDir=0):
		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="scrollScriptHelper", xDir = xDir, yDir = yDir)
		else:
			currentRect = libUSA.GetTrackingRectangle(next(iter(libUSA.magWindows)))
			currentRect = RectLTRB.fromCompatibleType(currentRect)
			displays = [wx.Display(i).GetGeometry() for i in range(wx.Display.GetCount())]
			width, height, (left, top) = mouseHandler.getTotalWidthAndHeightAndMinimumPosition(displays)
			screen = RectLTWH(left, top, width, height)
			magLevel = self.getSettings().magnificationLevel
			xSize = round(screen.right / magLevel)
			ySize = round(screen.bottom / magLevel)
			left, top, right, bottom = currentRect
			if xDir > 0:
				right = min(screen.right, currentRect.right + round(xDir * xSize / 2))
				left = right - xSize
			elif xDir < 0:
				left = max(screen.left, currentRect.left + round(xDir * xSize / 2))
				right = left + xSize
			if yDir > 0:
				bottom = min(screen.bottom, currentRect.bottom + round(yDir * ySize / 2))
				top = bottom - ySize
			elif yDir < 0:
				top = max(screen.top, currentRect.top + round(yDir * ySize / 2))
				bottom = top + ySize
			newRect = RectLTRB(left, top, right, bottom)
			if newRect == currentRect:
				ui.message(_("Edge of screen"))
				return
			self.trackToRectangle(newRect)

	def hookOurMouseFunction(self):
		# this function replaces the mousehook in NVDA with our own mousehook. we cannot use the mouseCallback since we
		# wont get the mousedata that we need (whether the scrollwheel goes up or down)
		global oldMouseHook
		oldMouseHook = winInputHook.mouseHook
		winInputHook.mouseHook = mouseHook2

		wx.CallLater(200, resetWinInputHook)

	def unhookOurMouseFunction(self):
		global oldMouseHook
		winInputHook.mouseHook = oldMouseHook

		wx.CallLater(200, resetWinInputHook)

	@script(gestures=["kb:alt+NVDA+-", "kb:control+pageDown", "kb:control+-"], description=_("Decreases the magnification level"))
	def script_decreaseMagnification(self, gesture):
		self._setMagnificationLevelHelper(increase=False)

	# kb:alt+NVDA+shift+0 (plus on english keyboard) and alt+NVDA+plus is for german keyboard users
	@script(gestures=["kb:alt+NVDA+=", "kb:control+pageUp", "kb:control+="], description=_("Increases the magnification level"))
	def script_increaseMagnification(self, gesture):
		self._setMagnificationLevelHelper(increase=True)

	def toggleMagnifier(self):

		if self.getSettings().isRemote:
			self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(command="toggleMagnifier")
		else:
			settings = self.getSettings()
			enabled = settings._toggleMagnifierHelper()
			if enabled:
				levelString = "{:.1f}".format(settings.magnificationLevel )
				ui.message(_("Magnifier on, level {level}").format(level=levelString))
			else:
				ui.message(_("Magnifier off"))

	@script(gestures=["kb:alt+NVDA+0", "kb:control+0"], description=_("Toggles the magnifier"))
	def script_toggleMagnifier(self, gesture):
		self.toggleMagnifier()

	def getDialogs(self):
		from ._base import dialogs
		return [dialogs.USACustomColorSchemeDialog, dialogs.USAFGColorSchemeDialog]


	@script(description= _("stop speech and browsemode if applicable."))
	def script_stopSpeechAndBrowseMode(self, gesture):
		log.debugWarning("script_stopSpeechAndBrowseMode called")
		if not self.getSettings().objectReaderMode:
			gesture.send()
			return
		# cancel speech as default
		speech.speech.cancelSpeech()

		# turn speech off when relevant
		if not speech.speech._speechState.speechMode == speech.speech.SpeechMode.off:
			log.debugWarning("turning off speech by script_stopSpeechAndBrowseMode")
			speech.speech.setSpeechMode(speech.speech.SpeechMode.off)
			return

		# turn off browse mode
		nav = api.getNavigatorObject()
		if nav.treeInterceptor:
			if isinstance(nav.treeInterceptor, browseMode.BrowseModeTreeInterceptor):
				if not nav.treeInterceptor.passThrough:
					log.debugWarning("turning off browseMode by script_stopSpeechAndBrowseMode")
					nav.treeInterceptor.passThrough = True
					nav.treeInterceptor.disableAutoPassThrough = True
					return

		gesture.send()


	@script(gestures=["kb:NVDA+downArrow",], description= _("sayall command in USA"))
	def script_sayAllUSA(self, gesture):
		log.info("our own Sayall is called!")
		speech.speech.setSpeechMode(speech.speech.SpeechMode.talk)
		globalCommands.commands.script_sayAll(gesture)


# get the NVDA keys beforehand toe make the code in the hook just a bit quicker.
NVDAKeys = keyboardHandler.getNVDAModifierKeys()
def resetWinInputHook():
	# only when hookThreadRefCount is 0, it makes sense to initialize wnnInputHook
	while winInputHook.hookThreadRefCount > 0:
		log.debugWarning(f"USA: terminating hookthread. current amount: {winInputHook.hookThreadRefCount}")
		winInputHook.terminate()

	log.debugWarning(f"USA: re-initialize winInputHook")
	winInputHook.initialize()

# function used in mouseHook to check if an NVDA key is pressed
def checkIfNVDAKeyIsPressed():
	global NVDAKeys
	for ii in NVDAKeys:
		# the NVDAKeys and currentModifiers are tuples with the first element as vkCode
		if ii[0] in [kk[0] for kk in keyboardHandler.currentModifiers]: return True
	return False

# windows hook functie om het scroll wiel event op te vangen en USA in te laten zoomen als een toets + scroll wheel
# ingedrukt wordt. Deze functie werkt niet perfect in python en het kan zijn dat het af en toe de toetscombi + scrollen
# niet afvangt. Dit is een kopie met wat andere functies toegevoegd van de winInputHook variant die ook in NVDA zit.
@ctypes.WINFUNCTYPE(ctypes.c_long, ctypes.c_int, ctypes.wintypes.WPARAM, ctypes.wintypes.LPARAM)
def mouseHook2( code, wParam, lParam):
	global USA_Instance
	try:
		HC_ACTION = 0
		# als NVDA vastloopt dan moet dit gewoon doorgaan
		if watchdog.isAttemptingRecovery or code != HC_ACTION:
			return oldMouseHook(code, wParam, lParam)

		if not checkIfNVDAKeyIsPressed():
			return oldMouseHook(code, wParam, lParam)

		# controleert off dit een mousewheelevent is of clickevent
		if not wParam in (522, 514, 513) :
			return oldMouseHook(code, wParam, lParam)
		USA_Instance = findUSAInstance()
		if not USA_Instance:
			return oldMouseHook(code, wParam, lParam)

		if wParam == 522:
			msll = winInputHook.MSLLHOOKSTRUCT.from_address(lParam)
			if int(msll.mouseData) == 7864320:
				wx.CallAfter(scrollZoom, increase=True)
				return 1
			elif int(msll.mouseData) == 4287102976:
				wx.CallAfter(scrollZoom, increase=False)
				return 1
		if wParam == 513:
			return 1
		if wParam == 514:
			log.debugWarning(f"USA: mouseHook2; should execute callReadAtMousePos")
			wx.CallLater(200, callReadAtMousePos)
			return 1

	except Exception as e:
		return oldMouseHook(code, wParam, lParam)

	# sometimes the code gets here and I do not know how. Maybe because of the slowness of python. Just return 1 here
	# to prevent scroll events from going through when we do not want to
	return 1


def scrollZoom(increase = False):
	# deze functie is appart van de _setMagnificaitonLevelHelper vanwege wat problemen met wx.
	global USA_Instance
	USA_Instance = findUSAInstance()
	log.debugWarning(f"USA: scrollZoom; call {USA_Instance} method with increase {increase}")
	USA_Instance._setMagnificationLevelHelper(increase)

def callReadAtMousePos():
	global USA_Instance
	USA_Instance = findUSAInstance()
	log.debugWarning(f"USA: callReadAtMousePos; call {USA_Instance} method to read from mouse pos.")
	wx.CallAfter(USA_Instance.script_readAtMousePos, None)

USA_Instance = None

def findUSAInstance():
	global USA_Instance
	if USA_Instance: return USA_Instance

	try:
		import vision
		instances = vision.handler.getActiveProviderInstances()
		for ii in instances:
			if ii.getSettings().getId().lower().strip() == "usa":
				return ii
	except:
		pass
	log.error("USA: findUSAInstance; could not find USA instance.")
	return None


VisionEnhancementProvider = USA
